@@ -21,7 +21,7 @@ import Data.ByteString.Lazy qualified as LBS
 import Data.Char
 import Data.Functor (($>))
 import Data.Int (Int64)
-import Data.List (foldl', transpose)
+import Data.List qualified as L
 import Data.Map qualified as M
 import Data.Maybe
 import Data.Set qualified as S
@@ -537,7 +537,7 @@ system prog options input = do
     prog' = "'" <> T.pack prog <> "'"
 
 formatDataForGnuplot :: [Value] -> T.Text
-formatDataForGnuplot = T.unlines . map line . transpose . map valueElems
+formatDataForGnuplot = T.unlines . map line . L.transpose . map valueElems
   where
     line = T.unwords . map prettyText
 
@@ -670,9 +670,12 @@ data Options = Options
     scriptSkipCompilation :: Bool,
     scriptOutput :: Maybe FilePath,
     scriptVerbose :: Int,
-    scriptStopOnError :: Bool
+    scriptStopOnError :: Bool,
+    scriptBinary :: Bool,
+    scriptExps :: [Either FilePath T.Text]
   }
 
+-- | The configuration before any user-provided options are processed.
 initialOptions :: Options
 initialOptions =
   Options
@@ -683,7 +686,9 @@ initialOptions =
       scriptSkipCompilation = False,
       scriptOutput = Nothing,
       scriptVerbose = 0,
-      scriptStopOnError = False
+      scriptStopOnError = False,
+      scriptBinary = False,
+      scriptExps = []
     }
 
 data Env = Env
@@ -1070,8 +1075,9 @@ processScript env script = do
   (failures, outputs, files) <-
     unzip3 <$> mapM (processBlock env) script
   cleanupImgDir env $ mconcat files
-  pure (foldl' min Success failures, T.intercalate "\n" outputs)
+  pure (L.foldl' min Success failures, T.intercalate "\n" outputs)
 
+-- | Common command line options that transform 'Options'.
 scriptCommandLineOptions :: [FunOptDescr Options]
 scriptCommandLineOptions =
   [ Option
@@ -1139,6 +1145,9 @@ commandLineOptions =
            "Stop and do not produce output file if any directive fails."
        ]
 
+-- | Start up (and eventually shut down) a Futhark server
+-- corresponding to the provided program. If the program has a @.fut@
+-- extension, it will be compiled automatically.
 prepareServer :: FilePath -> Options -> (ScriptServer -> IO a) -> IO a
 prepareServer prog opts f = do
   futhark <- maybe getExecutablePath pure $ scriptFuthark opts
@@ -1156,7 +1165,7 @@ prepareServer prog opts f = do
           unwords compile_options
 
     let onError err = do
-          mapM_ (T.hPutStrLn stderr) err
+          T.hPutStrLn stderr err
           exitFailure
 
     void $
