@@ -498,6 +498,12 @@ newtype ErrorMsg a = ErrorMsg [ErrorMsgPart a]
 instance IsString (ErrorMsg a) where
   fromString = ErrorMsg . pure . fromString
 
+instance Monoid (ErrorMsg a) where
+  mempty = ErrorMsg mempty
+
+instance Semigroup (ErrorMsg a) where
+  ErrorMsg x <> ErrorMsg y = ErrorMsg $ x <> y
+
 -- | A part of an error message.
 data ErrorMsgPart a
   = -- | A literal string.
@@ -608,6 +614,11 @@ data OpaqueType
     -- represent that constructor payload. This is necessary because
     -- we deduplicate payloads across constructors.
     OpaqueSum [ValueType] [(Name, [(EntryPointType, [Int])])]
+  | -- | An array with this rank and named opaque element type.
+    OpaqueArray Int Name [ValueType]
+  | -- | An array with known rank and where the elements are this
+    -- record type.
+    OpaqueRecordArray Int Name [(Name, EntryPointType)]
   deriving (Eq, Ord, Show)
 
 -- | Names of opaque types and their representation.
