@@ -236,20 +236,8 @@ calculateDims :: Exp -> ExpReplacements -> MonoM Exp
 calculateDims body repl =
   foldCalc top_repl $ expReplace top_repl body
   where
-    -- list of strict sub-expressions of e
-    subExps e
-      | Just e' <- stripExp e = subExps e'
-      | otherwise = astMap mapper e `execState` mempty
-      where
-        mapOnExp e'
-          | Just e'' <- stripExp e' = mapOnExp e''
-          | otherwise = do
-              modify (ReplacedExp e' :)
-              astMap mapper e'
-        mapper = identityMapper {mapOnExp}
-    depends (a, _) (b, _) = b `elem` subExps (unReplaced a)
-    top_repl =
-      topologicalSort depends repl
+    depends (a, _) (b, _) = unReplaced b `elem` subExps (unReplaced a)
+    top_repl = topologicalSort depends repl
 
     ---- Calculus insertion
     foldCalc [] body' = pure body'
@@ -513,8 +501,8 @@ transformAppExp (Apply fe args _) res =
     <*> transformAppRes res
   where
     onArg (Info ext, e) = (ext,) <$> transformExp e
-transformAppExp (Loop sparams pat e1 form body loc) res = do
-  e1' <- transformExp e1
+transformAppExp (Loop sparams pat loopinit form body loc) res = do
+  e1' <- transformExp $ loopInitExp loopinit
 
   let dimArgs = S.fromList sparams
   pat' <- withArgs dimArgs $ transformPat pat
@@ -540,7 +528,7 @@ transformAppExp (Loop sparams pat e1 form body loc) res = do
   -- sizes for them.
   (pat_sizes, pat'') <- sizesForPat pat'
   res' <- transformAppRes res
-  pure $ AppExp (Loop (sparams' ++ pat_sizes) pat'' e1' form' body' loc) (Info res')
+  pure $ AppExp (Loop (sparams' ++ pat_sizes) pat'' (LoopInitExplicit e1') form' body' loc) (Info res')
 transformAppExp (BinOp (fname, _) (Info t) (e1, d1) (e2, d2) loc) res = do
   (AppRes ret ext) <- transformAppRes res
   fname' <- transformFName loc fname (toStruct t)
@@ -633,11 +621,11 @@ transformExp (RecordLit fs loc) =
   where
     transformField (RecordFieldExplicit name e loc') =
       RecordFieldExplicit name <$> transformExp e <*> pure loc'
-    transformField (RecordFieldImplicit v t _) = do
+    transformField (RecordFieldImplicit (L vloc v) t _) = do
       t' <- traverse transformType t
       transformField $
         RecordFieldExplicit
-          (baseName v)
+          (L vloc (baseName v))
           (Var (qualName v) t' loc)
           loc
 transformExp (ArrayVal vs t loc) =
