@@ -596,7 +596,7 @@ worthIntrablock lam = bodyInterest (lambdaBody lam) > 1
             max
             (bodyInterest defbody)
             (map (bodyInterest . caseBody) cases)
-      | Op (Screma w _ (ScremaForm _ _ lam')) <- stmExp stm =
+      | Op (Screma w _ (ScremaForm lam' _ _)) <- stmExp stm =
           zeroIfTooSmall w + bodyInterest (lambdaBody lam')
       | Op (Stream _ _ _ lam') <- stmExp stm =
           bodyInterest $ lambdaBody lam'
@@ -625,7 +625,7 @@ worthSequentialising lam = bodyInterest (0 :: Int) (lambdaBody lam) > 1
     interest depth stm
       | "sequential" `inAttrs` attrs =
           0 :: Int
-      | Op (Screma _ _ form@(ScremaForm _ _ lam')) <- stmExp stm,
+      | Op (Screma _ _ form@(ScremaForm lam' _ _)) <- stmExp stm,
         isJust $ isMapSOAC form =
           if sequential_inner
             then 0
@@ -636,7 +636,7 @@ worthSequentialising lam = bodyInterest (0 :: Int) (lambdaBody lam) > 1
           bodyInterest (depth + 1) body * 10
       | WithAcc _ withacc_lam <- stmExp stm =
           bodyInterest (depth + 1) (lambdaBody withacc_lam)
-      | Op (Screma _ _ form@(ScremaForm _ _ lam')) <- stmExp stm =
+      | Op (Screma _ _ form@(ScremaForm lam' _ _)) <- stmExp stm =
           1
             + bodyInterest (depth + 1) (lambdaBody lam')
             +
